import os
import json
import numpy as np
import pandas as pd
import random
import unittest
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

from .. import ErrorAnalyzer
from .. import compute_accuracy_score, balanced_accuracy_score, compute_primary_model_accuracy, compute_confidence_decision

default_seed = 10
np.random.seed(default_seed)
random.seed(default_seed)

DATASET_URL = 'https://www.openml.org/data/get_csv/54002/adult-census.arff'
df = pd.read_csv(DATASET_URL)
target = 'class'

script_dir = os.path.dirname(__file__)
reference_result_file_path = os.path.join(script_dir, 'reference_data.json')
with open(reference_result_file_path) as f:
    reference_data = json.load(f)

numeric_features = df.select_dtypes(include=['int64', 'float64']).columns.tolist()
categorical_features = df.select_dtypes(include=['object']).drop([target], axis=1).columns.tolist()
X_numerical = df.dropna().drop(target, axis=1)[numeric_features]
X_all = df.dropna().drop(target, axis=1)
y = df.dropna()[target]


class TestErrorAnalyzer(unittest.TestCase):

    def test_with_only_scikit_model(self):
        X_train, X_test, y_train, y_test = train_test_split(X_numerical, y, test_size=0.2)
        rf = RandomForestClassifier(n_estimators=10)
        rf.fit(X_train, y_train)
        error_tree = ErrorAnalyzer(rf, feature_names=numeric_features)
        error_tree.fit(X_test, y_test)

        y_true, _ = error_tree._compute_primary_model_error(X_test.values, y_test)
        y_pred = error_tree._error_tree.estimator_.predict(X_test.values)

        error_tree_accuracy_score = compute_accuracy_score(y_true, y_pred)
        error_tree_balanced_accuracy = balanced_accuracy_score(y_true, y_pred)
        primary_model_predicted_accuracy = compute_primary_model_accuracy(y_pred)
        primary_model_true_accuracy = compute_primary_model_accuracy(y_true)
        fidelity, confidence_decision = compute_confidence_decision(primary_model_true_accuracy,
                                                                    primary_model_predicted_accuracy)

        metric_to_check = {
            'error_tree_accuracy_score': error_tree_accuracy_score,
            'error_tree_balanced_accuracy': error_tree_balanced_accuracy,
            'primary_model_predicted_accuracy': primary_model_predicted_accuracy,
            'primary_model_true_accuracy': primary_model_true_accuracy,
            'fidelity': fidelity
        }

        reference_data_for_single_estimator = reference_data.get('single_estimator')
        metric_reference = reference_data_for_single_estimator.get('metric_reference')

        for metric_name, metric_value in metric_to_check.items():
            self.assertAlmostEqual(metric_value, metric_reference[metric_name], 5)

        leaf_summary_str = error_tree.get_error_leaf_summary(leaf_selector=98, output_format='str')
        leaf_summary_dct = error_tree.get_error_leaf_summary(leaf_selector=98, output_format='dict')
        single_leaf_summary_reference_dict = reference_data_for_single_estimator.get('single_leaf_summary_reference')
        self.assertListEqual(leaf_summary_str, single_leaf_summary_reference_dict['str'])
        self.assertListEqual(leaf_summary_dct, single_leaf_summary_reference_dict['dct'])

        all_leaves_summary_str = error_tree.get_error_leaf_summary(output_format='str')
        all_leaves_summary_dct = error_tree.get_error_leaf_summary(output_format='dict')
        all_leaves_summary_reference_dict = reference_data_for_single_estimator.get('all_leaves_summary_reference')
        self.assertListEqual(all_leaves_summary_str, all_leaves_summary_reference_dict['str'])
        self.assertListEqual(all_leaves_summary_dct, all_leaves_summary_reference_dict['dct'])

        evaluate_str = error_tree.evaluate(X_test, y_test, output_format='str')
        evaluate_dct = error_tree.evaluate(X_test, y_test, output_format='dict')
        evaluate_reference_dict = reference_data_for_single_estimator.get('evaluate_reference')
        self.assertEqual(evaluate_str, evaluate_reference_dict['str'])
        self.assertDictEqual(evaluate_dct, evaluate_reference_dict['dct'])


    def test_with_scikit_pipeline(self):
        X_train, X_test, y_train, y_test = train_test_split(X_all, y, test_size=0.2)
        numeric_transformer = Pipeline(steps=[
            ('SimpleImputer', SimpleImputer(strategy='median', add_indicator=True)),
            ('StandardScaler', StandardScaler()),
        ])
        categorical_transformer = Pipeline(steps=[
            ('OneHotEncoder', OneHotEncoder(handle_unknown='ignore')),
        ])
        preprocessor = ColumnTransformer(
            transformers=[
                ('num', numeric_transformer, numeric_features),
                ('cat', categorical_transformer, categorical_features)
            ])
        rf = Pipeline(steps=[('preprocessor', preprocessor),
                             ('classifier', RandomForestClassifier(n_estimators=10))])

        rf.fit(X_train, y_train)
        error_tree = ErrorAnalyzer(rf)
        error_tree.fit(X_test, y_test)

        X_test_prep, y_test_prep = error_tree.pipeline_preprocessor.transform(X_test), np.array(y_test)

        y_true, _ = error_tree._compute_primary_model_error(X_test_prep, y_test_prep)
        y_pred = error_tree._error_tree.estimator_.predict(X_test_prep)

        error_tree_accuracy_score = compute_accuracy_score(y_true, y_pred)
        error_tree_balanced_accuracy = balanced_accuracy_score(y_true, y_pred)
        primary_model_predicted_accuracy = compute_primary_model_accuracy(y_pred)
        primary_model_true_accuracy = compute_primary_model_accuracy(y_true)
        fidelity, confidence_decision = compute_confidence_decision(primary_model_true_accuracy,
                                                                    primary_model_predicted_accuracy)

        metric_to_check = {
            'error_tree_accuracy_score': error_tree_accuracy_score,
            'error_tree_balanced_accuracy': error_tree_balanced_accuracy,
            'primary_model_predicted_accuracy': primary_model_predicted_accuracy,
            'primary_model_true_accuracy': primary_model_true_accuracy,
            'fidelity': fidelity
        }

        reference_data_for_pipeline = reference_data.get('pipeline')
        metric_reference = reference_data_for_pipeline.get('metric_reference')

        for metric_name, metric_value in metric_to_check.items():
            self.assertAlmostEqual(metric_value, metric_reference[metric_name], 5)

        _ = error_tree.get_error_leaf_summary(leaf_selector=98, add_path_to_leaves=True, output_format='str')

        leaf_summary_str = error_tree.get_error_leaf_summary(leaf_selector=98, output_format='str')
        leaf_summary_dct = error_tree.get_error_leaf_summary(leaf_selector=98, output_format='dict')
        single_leaf_summary_reference_dict = reference_data_for_pipeline.get('single_leaf_summary_reference')
        self.assertListEqual(leaf_summary_str, single_leaf_summary_reference_dict['str'])
        self.assertListEqual(leaf_summary_dct, single_leaf_summary_reference_dict['dct'])

        all_leaves_summary_str = error_tree.get_error_leaf_summary(output_format='str')
        all_leaves_summary_dct = error_tree.get_error_leaf_summary(output_format='dict')
        all_leaves_summary_reference_dict = reference_data_for_pipeline.get('all_leaves_summary_reference')
        self.assertListEqual(all_leaves_summary_str, all_leaves_summary_reference_dict['str'])
        self.assertListEqual(all_leaves_summary_dct, all_leaves_summary_reference_dict['dct'])

        evaluate_str = error_tree.evaluate(X_test, y_test, output_format='str')
        evaluate_dct = error_tree.evaluate(X_test, y_test, output_format='dict')
        evaluate_reference_dict = reference_data_for_pipeline.get('evaluate_reference')
        self.assertEqual(evaluate_str, evaluate_reference_dict['str'])
        self.assertDictEqual(evaluate_dct, evaluate_reference_dict['dct'])



# -*- coding: utf-8 -*-
import numpy as np
from sklearn.exceptions import NotFittedError
from mealy.constants import ErrorAnalyzerConstants
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='mealy | %(levelname)s - %(message)s')

class ErrorTree(object):

    def __init__(self, error_decision_tree):

        self._estimator = error_decision_tree
        self._leaf_ids = None
        self._impurity = None
        self._quantized_impurity = None
        self._difference = None
        self._total_error_fraction = None
        self._error_class_idx = None
        self._wrongly_predicted_leaves = None
        self._correctly_predicted_leaves = None

        self._check_error_tree()

    @property
    def estimator_(self):
        if self._estimator is None:
            raise NotFittedError("You should fit the ErrorAnalyzer first")
        return self._estimator

    @property
    def impurity(self):
        if self._impurity is None:
            self._impurity = self.correctly_predicted_leaves / (self.wrongly_predicted_leaves + self.correctly_predicted_leaves)
        return self._impurity

    @property
    def quantized_impurity(self):
        if self._quantized_impurity is None:
            purity_bins = np.linspace(0, 1., ErrorAnalyzerConstants.NUMBER_PURITY_LEVELS)
            self._quantized_impurity = np.digitize(self.impurity, purity_bins)
        return self._quantized_impurity

    @property
    def difference(self):
        if self._difference is None:
            self._difference = self.correctly_predicted_leaves - self.wrongly_predicted_leaves  # only negative numbers
        return self._difference

    @property
    def total_error_fraction(self):
        if self._total_error_fraction is None:
            self._total_error_fraction = self.wrongly_predicted_leaves / self.n_total_errors
        return self._total_error_fraction

    @property
    def error_class_idx(self):
        if self._error_class_idx is None:
            self._error_class_idx = np.where(self.estimator_.classes_ == ErrorAnalyzerConstants.WRONG_PREDICTION)[0][0]
        return self._error_class_idx

    @property
    def n_total_errors(self):
        return self.estimator_.tree_.value[0, 0, self.error_class_idx]

    @property
    def wrongly_predicted_leaves(self):
        if self._wrongly_predicted_leaves is None:
            self._wrongly_predicted_leaves = self.estimator_.tree_.value[self.leaf_ids, 0, self.error_class_idx]
        return self._wrongly_predicted_leaves

    @property
    def correctly_predicted_leaves(self):
        if self._correctly_predicted_leaves is None:
            self._correctly_predicted_leaves = self.estimator_.tree_.value[self.leaf_ids, 0, 1 - self.error_class_idx]
        return self._correctly_predicted_leaves

    @property
    def leaf_ids(self):
        if self._leaf_ids is None:
            self._leaf_ids = np.where(self.estimator_.tree_.feature < 0)[0]
        return self._leaf_ids

    def _check_error_tree(self):
        if self.estimator_.tree_.node_count == 1:
            logger.warning("The error tree has only one node, there will be problems when using it with ErrorVisualizer")



# Dataiku Model Error Analysis lib
from mealy.constants import ErrorAnalyzerConstants
from mealy.error_analyzer import ErrorAnalyzer
from mealy.preprocessing import PipelinePreprocessor, DummyPipelinePreprocessor
from mealy.error_visualizer import ErrorVisualizer, _BaseErrorVisualizer
from mealy.error_tree import ErrorTree
from mealy.metrics import compute_accuracy_score, balanced_accuracy_score, compute_primary_model_accuracy, compute_confidence_decision
from mealy.error_analysis_utils import generate_preprocessing_steps



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from scipy.sparse import issparse
from collections import defaultdict
import logging
from mealy.error_analysis_utils import check_lists_having_same_elements, generate_preprocessing_steps, invert_transform_via_identity
from mealy.constants import ErrorAnalyzerConstants

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='mealy | %(levelname)s - %(message)s')


class FeatureNameTransformer(object):
    """ Transformer of feature names and indices.

        A FeatureNameTransformer parses an input Pipeline preprocessor and generate
        a mapping between the input unprocessed feature names/indices and the output
        preprocessed feature names/indices.

        Args:
            ct_preprocessor (sklearn.compose.ColumnTransformer): preprocessor.
            orig_feats (list): list of original unpreprocessed feature names, default=None.

        Attributes:
            original_feature_names (list): list of original unpreprocessed feature names.
            preprocessed_feature_names (list): list of preprocessed feature names.

    """
    def __init__(self, original_features, preprocessed_features):
        self.original_feature_names = original_features
        self.preprocessed_feature_names = preprocessed_features

    def get_original_feature_names(self):
        return self.original_feature_names

    def get_preprocessed_feature_names(self):
        return self.preprocessed_feature_names

    def is_categorical(self, index=None, name=None):
        raise NotImplementedError

    def inverse_transform_feature_id(self, index):
        raise NotImplementedError

    def inverse_transform(self, x):
        raise NotImplementedError

    def transform(self, x):
        raise NotImplementedError

    def get_top_ranked_feature_ids(self, feature_importances, max_nr_features):
        raise NotImplementedError

    def inverse_thresholds(self, tree):
        raise NotImplementedError


class PipelinePreprocessor(FeatureNameTransformer):
    """Transformer of feature values from the original values to preprocessed ones.

        A PipelinePreprocessor parses an input Pipeline preprocessor and generate
        a mapping between the input unprocessed feature values and the output
        preprocessed feature values.

        Args:
            ct_preprocessor (sklearn.compose.ColumnTransformer): preprocessing steps.
            original_features (list): list of original unpreprocessed feature names, default=None.

    """

    def __init__(self, ct_preprocessor, original_features=None):
        self.ct_preprocessor = ct_preprocessor
        self.original2preprocessed = defaultdict(list)
        self.preprocessed2original = {}
        self.categorical_features = []

        logger.info('Retrieving the list of features used in the pipeline')
        original_features_from_ct = self._get_feature_list_from_column_transformer()
        if original_features is None:
            original_features = original_features_from_ct
        elif not check_lists_having_same_elements(original_features, original_features_from_ct):
            # If user explicitly gives a list of input features, we compare it with the list derived from the ColumnTransformer
            raise ValueError('The list of features given by user does not correspond to the list of features handled by the Pipeline.')

        super(PipelinePreprocessor, self).__init__(original_features=original_features, preprocessed_features=[])

        logger.info('Generating the feature id mapping dict')
        self._create_feature_mapping()

    def _get_feature_list_from_column_transformer(self):
        all_features = []
        for _, transformer, feature_names in self.ct_preprocessor.transformers_:
            for step in generate_preprocessing_steps(transformer):
                if isinstance(step, ErrorAnalyzerConstants.VALID_CATEGORICAL_STEPS):
                    # Check for categorical features
                    self.categorical_features += feature_names
                    break

            all_features += feature_names
        return all_features

    def _create_feature_mapping(self):
        """
        Update the dicts of input <-> output feature id mapping: self.original2preprocessed and self.preprocessed2original
        """
        for _, transformer, feature_names in self.ct_preprocessor.transformers_:
            orig_feat_ids = np.where(np.in1d(self.original_feature_names, feature_names))[0]
            for step in generate_preprocessing_steps(transformer):
                output_dim_changed = False
                if isinstance(step, ErrorAnalyzerConstants.STEPS_THAT_CHANGE_OUTPUT_DIMENSION_WITH_OUTPUT_FEATURE_NAMES):
                    # It is assumed that for each pipeline, at most one step changes the feature's dimension
                    # For now, it can only be a OneHotEncoder step
                    self._update_feature_mapping_dict_using_output_names(step,
                                                                        feature_names,
                                                                        orig_feat_ids)
                    output_dim_changed = True
                    break
            if not output_dim_changed:
                self._update_feature_mapping_dict_using_input_names(feature_names, orig_feat_ids)

    def _update_feature_mapping_dict_using_input_names(self, transformer_feature_names, original_feature_ids):
        self.preprocessed_feature_names.extend(transformer_feature_names)
        for original_feat_id in original_feature_ids:
            idx = len(self.preprocessed2original)
            self.original2preprocessed[original_feat_id] = [idx]
            self.preprocessed2original[idx] = original_feat_id

    def _update_feature_mapping_dict_using_output_names(self, single_transformer, transformer_feature_names, original_feature_ids):
        out_feature_names = list(single_transformer.get_feature_names(input_features=transformer_feature_names))
        self.preprocessed_feature_names.extend(out_feature_names)
        for orig_id, orig_name in zip(original_feature_ids, transformer_feature_names):
            part_out_feature_names = [name for name in out_feature_names if orig_name + '_' in name]
            offset = len(self.preprocessed2original)
            for i in range(len(part_out_feature_names)):
                self.original2preprocessed[orig_id].append(offset + i)
                self.preprocessed2original[offset + i] = orig_id

    def _transform_feature_id(self, index):
        """
        Args:
            index: int

        Returns: index of output feature(s) generated by the requested feature.
        """
        return self.original2preprocessed[index]

    def transform(self, x):
        """Transform the input feature values according to the preprocessing pipeline.

        Args:
            x (array-like or dataframe of shape (number of samples, number of features)): input feature values.

        Return:
            numpy.ndarray: transformed feature values.
        """
        return self.ct_preprocessor.transform(x)

    def _get_feature_ids_related_to_transformer(self, transformer_feature_names):
        original_features = self.get_original_feature_names()
        original_feature_ids = np.where(np.in1d(original_features, transformer_feature_names))[0]
        preprocessed_feature_ids = []
        for i in original_feature_ids:
            preprocessed_feature_ids += self._transform_feature_id(i)
        return original_feature_ids, preprocessed_feature_ids

    @staticmethod
    def _inverse_single_step(single_step, step_output, transformer_feature_names):
        inverse_transform_function_available = getattr(single_step, "inverse_transform", None)
        if invert_transform_via_identity(single_step):
            logger.info("Reversing step using identity transformation on column(s): {}".format(single_step, ', '.join(transformer_feature_names)))
            return step_output
        if inverse_transform_function_available:
            logger.info("Reversing step using inverse_transform() method on column(s): {}".format(single_step, ', '.join(transformer_feature_names)))
            return single_step.inverse_transform(step_output)
        raise TypeError('The package does not support {} because it does not provide inverse_transform function.'.format(single_step))

    def inverse_transform(self, preprocessed_x):
        """Invert the preprocessing pipeline and inverse transform feature values.

        Args:
            preprocessed_x (numpy.ndarray or scipy sparse matrix): preprocessed feature values.

        Return:
            numpy.ndarray: feature values without preprocessing.

        """
        nr_original_features = len(self.get_original_feature_names())
        undo_prep_test_x = np.zeros((preprocessed_x.shape[0], nr_original_features), dtype='O')
        any_cat = np.vectorize(lambda x: self.is_categorical(x))

        for _, transformer, feature_names in reversed(self.ct_preprocessor.transformers_):
            original_feature_ids, preprocessed_feature_ids = self._get_feature_ids_related_to_transformer(feature_names)
            transformer_output = preprocessed_x[:, preprocessed_feature_ids]
            if issparse(transformer_output) and not np.any(any_cat(original_feature_ids)):
                transformer_output = transformer_output.todense()

            # TODO: could be simplified as sklearn.Pipeline implements inverse_transform
            for step in generate_preprocessing_steps(transformer, invert_order=True):
                transformer_input = PipelinePreprocessor._inverse_single_step(step, transformer_output, feature_names)
                transformer_output = transformer_input
            undo_prep_test_x[:, original_feature_ids] = transformer_input

        return undo_prep_test_x

    def is_categorical(self, index=None, name=None):
        """Check whether an unprocessed feature at a given index or with a given name is categorical.

        Args:
            index (int): feature index.
            name (str): feature name.

        Return:
            bool: True if the input feature is categorical, else False. If both index and name are provided, the index
                is retained.
        """
        if index is not None:
            name = self.original_feature_names[index]
        if name is not None:
            return name in self.categorical_features
        else:
            raise ValueError("Either the input index or its name should be specified.")

    def inverse_transform_feature_id(self, index):
        """Undo preprocessing of feature name.

        Transform the preprocessed feature name at given index back into the original unprocessed feature index.

        Args:
            index (int): feature index.

        Return:
            int : index of the unprocessed feature corresponding to the input preprocessed feature index.
        """
        return self.preprocessed2original[index]

    def get_top_ranked_feature_ids(self, feature_importances, max_nr_features):
        ranked_transformed_feature_ids = np.argsort(- feature_importances)
        if max_nr_features <= 0:
            max_nr_features += len(self.get_original_feature_names())

        ranked_feature_ids, seen = [], set()
        for idx in ranked_transformed_feature_ids:
            inverse_transformed_feature_id = self.inverse_transform_feature_id(idx)
            if inverse_transformed_feature_id not in seen:
                seen.add(inverse_transformed_feature_id)
                ranked_feature_ids.append(inverse_transformed_feature_id)
                if max_nr_features == len(ranked_feature_ids):
                    return ranked_feature_ids
        return ranked_feature_ids # should never be reached, but just in case

    def inverse_thresholds(self, tree):
        used_feature_mask = tree.feature >= 0
        feats_idx = tree.feature[used_feature_mask]
        thresholds = tree.threshold.astype('O')
        thresh = thresholds[used_feature_mask]
        n_cols = len(self.get_preprocessed_feature_names())

        dummy_x, indices= [], []
        for f, t in zip(feats_idx, thresh):
            row = [0]*n_cols
            row[f] = t
            dummy_x.append(row)
            indices.append(self.inverse_transform_feature_id(f))

        undo_dummy_x = self.inverse_transform(np.array(dummy_x))
        descaled_thresh = [undo_dummy_x[i, j] for i, j in enumerate(indices)]
        thresholds[used_feature_mask] = descaled_thresh
        return thresholds


class DummyPipelinePreprocessor(FeatureNameTransformer):

    def __init__(self, model_performance_predictor_features):
        super(DummyPipelinePreprocessor, self).__init__(
            original_features=model_performance_predictor_features,
            preprocessed_features=model_performance_predictor_features)

    def transform(self, x):
        """
        Args:
            x (array-like or dataframe of shape (number of samples, number of features)): input feature values.
        Returns:
            ndarray
        """
        if isinstance(x, pd.DataFrame):
            return x.values
        if isinstance(x, np.ndarray) or issparse(x):
            return x
        raise TypeError('x should be either a pandas dataframe, a numpy ndarray or a scipy sparse matrix')

    def is_categorical(self, index=None, name=None):
        return False

    def inverse_transform_feature_id(self, index):
        return index

    def inverse_transform(self, x):
        return x

    def get_top_ranked_feature_ids(self, feature_importances, max_nr_features):
        if max_nr_features == 0:
            return np.argsort(- feature_importances)
        return np.argsort(- feature_importances)[:max_nr_features]

    def inverse_thresholds(self, tree):
        return tree.threshold.astype('O')



import numpy as np
import pandas as pd
import logging
from unittest import TestCase, skip
from unittest.mock import patch, Mock
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.tree import DecisionTreeClassifier

from .. import ErrorTree, ErrorAnalyzer, ErrorAnalyzerConstants, DummyPipelinePreprocessor, PipelinePreprocessor

logger = logging.getLogger(__name__)

class TestErrorAnalyzer(TestCase):
    @patch("numpy.count_nonzero", wraps=np.count_nonzero)
    def test_get_leaf_selector(self, cnz):
        leaves = [4, 80, 13, 3, 1, 0, 12]
        selector = ErrorAnalyzer._get_leaf_selector(list(leaves))
        self.assertFalse(cnz.called)
        self.assertTrue(selector(True))
        np.testing.assert_array_equal(selector(leaves), leaves)

        cnz.reset_mock()
        selector = ErrorAnalyzer._get_leaf_selector(list(leaves), [3, 4])
        cnz.assert_called_once()
        np.testing.assert_array_equal(cnz.call_args[0][0],
            [True, False, False, True, False, False, False])
        np.testing.assert_array_equal(selector(np.array(leaves)), [4, 3])

        cnz.reset_mock()
        selector = ErrorAnalyzer._get_leaf_selector(list(leaves), 1)
        cnz.assert_called_once()
        np.testing.assert_array_equal(cnz.call_args[0][0],
            [False, False, False, False, True, False, False])
        np.testing.assert_array_equal(selector(np.array(leaves)), [1])

        cnz.reset_mock()
        with self.assertLogs("mealy.error_analyzer") as caplog:
            selector = ErrorAnalyzer._get_leaf_selector(list(leaves), -1)
        self.assertEqual(len(caplog.records), 1)
        self.assertEqual(caplog.records[-1].msg,
            "None of the ids provided correspond to a leaf id.")
        cnz.assert_called_once()
        np.testing.assert_array_equal(cnz.call_args[0][0],
            [False, False, False, False, False, False, False])
        self.assertEqual(selector(np.array(leaves)).size, 0)

        cnz.reset_mock()
        with self.assertLogs("mealy.error_analyzer") as caplog:
            selector = ErrorAnalyzer._get_leaf_selector(list(leaves), [100, 0, -1])
        self.assertEqual(len(caplog.records), 1)
        self.assertEqual(caplog.records[-1].msg,
            "Some of the ids provided do not belong to leaves. Only leaf ids are kept.")
        cnz.assert_called_once()
        np.testing.assert_array_equal(cnz.call_args[0][0],
            [False, False, False, False, False, True, False])
        np.testing.assert_array_equal(selector(np.array(leaves)), [0])

    def test_init(self):
        # Analyzer with single estimator
        single_estimator = Mock(spec=BaseEstimator)
        with patch("mealy.error_analyzer.DummyPipelinePreprocessor.__init__", return_value=None) as dummy:
            analyzer = ErrorAnalyzer(single_estimator)
            dummy.assert_called_once_with(None)
        self.assertEqual(single_estimator, analyzer._primary_model)
        self.assertTrue(isinstance(analyzer.pipeline_preprocessor, DummyPipelinePreprocessor))

        # Analyzer with pipeline
        ct_transformer = Mock(spec=ColumnTransformer)
        ok_pipe = Mock(spec=Pipeline, steps=[
            ("transformer", ct_transformer),
            ("single_estimator", single_estimator)
        ])

        with patch("mealy.error_analyzer.PipelinePreprocessor.__init__", return_value=None) as pipe,\
            self.assertLogs() as caplog:
            analyzer = ErrorAnalyzer(ok_pipe)
            logger.info("Dummy log")
            pipe.assert_called_once_with(ct_transformer, None)
            self.assertEqual(len(caplog.records), 1)
            self.assertEqual(caplog.records[-1].msg, "Dummy log")
        self.assertEqual(single_estimator, analyzer._primary_model)
        self.assertTrue(isinstance(analyzer.pipeline_preprocessor, PipelinePreprocessor))

        # Analyzer with pipeline (wrong attribute type)
        wrong_primary_model = Mock()
        with self.assertRaisesRegex(TypeError, "ErrorAnalyzer needs as input either a scikit BaseEstimator or a scikit Pipeline."):
            analyzer = ErrorAnalyzer(wrong_primary_model)

        # Analyzer with pipeline (missing col transformer)
        incomplete_pipe = Mock(spec=Pipeline, steps=[
            ("single_estimator", single_estimator)
        ])

        with patch("mealy.error_analyzer.PipelinePreprocessor.__init__", return_value=None) as pipe,\
            self.assertLogs("mealy.error_analyzer", level="WARNING") as caplog,\
            self.assertRaisesRegex(TypeError, "The input preprocessor has to be a ColumnTransformer."):
            analyzer = ErrorAnalyzer(incomplete_pipe)
            self.assertEqual(len(caplog.records), 1)
            self.assertEqual(caplog.records[-1].msg,
                "Pipeline should have two steps: the preprocessing of the features, and the primary model to analyze.")

        # Analyzer with pipeline (missing estimator)
        incomplete_pipe = Mock(spec=Pipeline, steps=[
            ("wrong_type", Mock())
        ])

        with patch("mealy.error_analyzer.PipelinePreprocessor.__init__", return_value=None) as pipe,\
            self.assertLogs("mealy.error_analyzer", level="WARNING") as caplog,\
            self.assertRaisesRegex(TypeError, "The last step of the pipeline has to be a BaseEstimator"):
            analyzer = ErrorAnalyzer(incomplete_pipe)
            self.assertEqual(len(caplog.records), 1)
            self.assertEqual(caplog.records[-1].msg,
            "Pipeline should have two steps: the preprocessing of the features, and the primary model to analyze.")

    @staticmethod
    def get_analyzer(**kwargs):
        with patch("mealy.error_analyzer.ErrorAnalyzer.__init__", return_value=None):
            analyzer = ErrorAnalyzer(Mock(spec=BaseEstimator))
        error_tree = Mock(leaf_ids=np.array([4, 80, 13, 3, 1, 0, 12]), **kwargs)
        analyzer._error_tree = error_tree
        analyzer._random_state = 42
        analyzer._param_grid = None
        return analyzer

    def test_get_ranked_leaf_ids__tot_error_fraction(self):
        analyzer = TestErrorAnalyzer.get_analyzer(
            total_error_fraction=np.array([.4, .1, .2, .5, .6, .7, .3])
        )
        with patch.object(analyzer, "_get_leaf_selector", return_value=lambda array: array):
            sorted_leaves = analyzer._get_ranked_leaf_ids()
        np.testing.assert_array_equal(sorted_leaves, [0, 1, 3, 4, 12, 13, 80])

    def test_get_ranked_leaf_ids__purity(self):
        analyzer = TestErrorAnalyzer.get_analyzer(
            quantized_impurity=np.array([.5, .5, .5, .2, .1, .1, .5]),
            difference=np.array([0, 3, 2, 100, 11, 10, 1])
        )
        with patch.object(analyzer, "_get_leaf_selector", return_value=lambda array: array):
            sorted_leaves = analyzer._get_ranked_leaf_ids(rank_by='purity')
        np.testing.assert_array_equal(sorted_leaves, [0, 1, 3, 4, 12, 13, 80])

    def test_get_ranked_leaf_ids__class_difference(self):
        analyzer = TestErrorAnalyzer.get_analyzer(
            impurity=np.array([.5, -1, 1, .2, 1000, 1000, .8]),
            difference=np.array([0, 10, 0, 0, -5, -10, 0])
        )
        with patch.object(analyzer, "_get_leaf_selector", return_value=lambda array: array):
            sorted_leaves = analyzer._get_ranked_leaf_ids(rank_by='class_difference')
        np.testing.assert_array_equal(sorted_leaves, [0, 1, 3, 4, 12, 13, 80])

    def test_get_ranked_leaf_ids__invalid(self):
        analyzer = TestErrorAnalyzer.get_analyzer()
        with patch.object(analyzer, "_get_leaf_selector", return_value=lambda array: array),\
            self.assertRaisesRegex(ValueError,
            "Input argument for rank_by is invalid. Should be 'total_error_fraction', 'purity' or 'class_difference'"):
            sorted_leaves = analyzer._get_ranked_leaf_ids(rank_by='toto')

    def test_get_ranked_leaf_ids__empty(self):
        analyzer = TestErrorAnalyzer.get_analyzer()
        with patch.object(analyzer, "_get_leaf_selector", return_value=lambda array: np.array([])):
            sorted_leaves = analyzer._get_ranked_leaf_ids()
        self.assertEqual(sorted_leaves.size, 0)

    def test_fit__not_enough_data(self):
        analyzer = TestErrorAnalyzer.get_analyzer()
        analyzer.pipeline_preprocessor = Mock()
        analyzer.pipeline_preprocessor.transform.side_effect = lambda x: x
        X = np.array([[1,2,3,4,5]]*5)
        y = None
        with patch("mealy.error_analyzer.ErrorAnalyzer._compute_primary_model_error", side_effect=lambda a, b: (a, .3)),\
            self.assertRaises(ValueError):
            analyzer.fit(X, y)

    def test_fit__default_grid(self):
        analyzer = TestErrorAnalyzer.get_analyzer()
        analyzer.pipeline_preprocessor = Mock()
        analyzer.pipeline_preprocessor.transform.side_effect = lambda x: x
        self.get_analyzer.param_grid = None
        X = np.array([[1,2,3,4,5]]*100)
        y = np.array([1]*100)
        with patch("mealy.error_analyzer.ErrorAnalyzer._compute_primary_model_error", side_effect=lambda a, b: (b, .3)), \
            patch("mealy.error_analyzer.GridSearchCV") as patched:
            analyzer.fit(X, y)
            clf = patched.call_args[0][0]
            param_grid, cv = patched.call_args[1]["param_grid"], patched.call_args[1]["cv"]
        self.assertTrue(isinstance(clf, DecisionTreeClassifier))
        self.assertListEqual(param_grid["max_depth"], [5, 10])
        np.testing.assert_array_equal(param_grid["min_samples_leaf"],  np.linspace(.01/5, .01, 5))
        self.assertEqual(cv, 5)
        np.testing.assert_array_equal(analyzer._error_train_y, y)
        np.testing.assert_array_equal(analyzer._error_train_x, X)
        self.assertTrue(isinstance(analyzer.error_tree, ErrorTree))

    @skip(reason="Not done yet")
    def test_get_path_to_node(self):
        pass

    @skip(reason="Not done yet")
    def test_get_error_leaf_summary(self):
        pass


class TestErrorAnalyzerPrimaryModelMethods(TestCase):
    def setUp(self):
        with patch("mealy.error_analyzer.ErrorAnalyzer.__init__", return_value=None):
            self.analyzer = ErrorAnalyzer(Mock(spec=BaseEstimator))
        self.analyzer.epsilon = None
        self.analyzer._primary_model = Mock()

    def test_compute_primary_model_error__regressor(self):
        self.analyzer._primary_model._estimator_type = "regressor"
        self.analyzer._primary_model.predict.side_effect = lambda array: array+1
        X = np.array([1,1,1,1,1])
        y = np.array([0,0,0,0,1])
        with patch.object(self.analyzer, "_evaluate_primary_model_predictions",
            side_effect= lambda y_true, y_pred: (y_true, y_pred)):
            y_true, y_pred = self.analyzer._compute_primary_model_error(X, y)
        np.testing.assert_array_equal(y_true, y)
        np.testing.assert_array_equal(y_pred, [2,2,2,2,2])

    def test_compute_primary_model_error__multi(self):
        self.analyzer._primary_model.predict.side_effect = lambda array: array+1
        X = np.array([1,1,1,1,1])
        y = np.array([0,1,2,3,4])
        with patch.object(self.analyzer, "_evaluate_primary_model_predictions",
            side_effect= lambda y_true, y_pred: (y_true, y_pred)):
            y_true, y_pred = self.analyzer._compute_primary_model_error(X, y)
        np.testing.assert_array_equal(y_true, y)
        np.testing.assert_array_equal(y_pred, [2,2,2,2,2])

    def test_compute_primary_model_error__binary(self):
        self.analyzer._primary_model.classes_ = ["W", "C"]
        self.analyzer._primary_model.predict_proba.side_effect = lambda array: array
        self.analyzer.probability_threshold = .2
        X = np.array([
            [0, .1],
            [0, .8],
            [0, .1],
            [0, .8],
            [0, .8],
        ])
        y = np.array([0,1,1,0,0])
        with patch.object(self.analyzer, "_evaluate_primary_model_predictions",
            side_effect= lambda y_true, y_pred: (y_true, y_pred)):
            y_true, y_pred = self.analyzer._compute_primary_model_error(X, y)
        np.testing.assert_array_equal(y_true, y)
        np.testing.assert_array_equal(y_pred, ["W","C","W","C","C"])

    def test_evaluate_primary_model_predictions__regression(self):
        self.analyzer._primary_model._estimator_type = "regressor"
        self.analyzer.epsilon = .4
        y_true = np.array([1,.2,-.3,-.7,0])
        y_pred = np.array([.1,0,0,0.2,1])
        with patch("mealy.error_analyzer.get_epsilon", return_value="") as patched,\
            self.assertLogs("mealy.error_analyzer") as caplog:
            error_y, error_rate = self.analyzer._evaluate_primary_model_predictions(y_true, y_pred)
            self.assertFalse(patched.called)
            self.assertEqual(len(caplog.records), 1)
            self.assertEqual((caplog.records[-1].levelname, caplog.records[-1].msg),
                ("INFO", 'The primary model has an error rate of 0.6'))
        np.testing.assert_array_equal(error_y, [
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION
        ])
        np.testing.assert_array_equal(error_rate, .6)

    def test_evaluate_primary_model_predictions__classif(self):
        y_true = np.array(["A","B","A","C","A"])
        y_pred = np.array(["B","B","A","A","B"])
        with patch("mealy.error_analyzer.get_epsilon", return_value="") as patched,\
            self.assertLogs("mealy.error_analyzer") as caplog:
            error_y, error_rate = self.analyzer._evaluate_primary_model_predictions(y_true, y_pred)
            self.assertFalse(patched.called)
            self.assertEqual(len(caplog.records), 1)
            self.assertEqual((caplog.records[-1].levelname, caplog.records[-1].msg),
                ("INFO", 'The primary model has an error rate of 0.6'))
        np.testing.assert_array_equal(error_y, [
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION
        ])
        np.testing.assert_array_equal(error_rate, .6)

    def test_evaluate_primary_model_predictions__all_wrong(self):
        self.analyzer._primary_model._estimator_type = "regressor"
        y_true = np.array([1,.2,-.3,-.7,0])
        y_pred = np.array([.1,0,0,0.2,1])
        with patch("mealy.error_analyzer.get_epsilon", return_value=.1) as patched,\
            self.assertLogs("mealy.error_analyzer") as caplog:
            error_y, error_rate = self.analyzer._evaluate_primary_model_predictions(y_true, y_pred)
            patched.assert_called_once()
            self.assertEqual(len(caplog.records), 2)
            self.assertEqual((caplog.records[-1].levelname, caplog.records[-1].msg),
                ("INFO", 'The primary model has an error rate of 1'))
            self.assertEqual((caplog.records[-2].levelname, caplog.records[-2].msg),
                ("WARNING", 'All predictions are Wrong prediction. To build a proper ErrorAnalyzer decision tree we need both correct and incorrect predictions'))
        np.testing.assert_array_equal(error_y, [
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION
        ])
        np.testing.assert_array_equal(error_rate, 1)

    def test_evaluate_primary_model_predictions__all_correct(self):
        self.analyzer._primary_model._estimator_type = "regressor"
        y_true = np.array([1,.2,-.3,-.7,0])
        y_pred = np.array([.1,0,0,0.2,1])
        with patch("mealy.error_analyzer.get_epsilon", return_value=1.1) as patched,\
            self.assertLogs("mealy.error_analyzer") as caplog:
            error_y, error_rate = self.analyzer._evaluate_primary_model_predictions(y_true, y_pred)
            patched.assert_called_once()
            self.assertEqual(len(caplog.records), 2)
            self.assertEqual((caplog.records[-1].levelname, caplog.records[-1].msg),
                ("INFO", 'The primary model has an error rate of 0'))
            self.assertEqual((caplog.records[-2].levelname, caplog.records[-2].msg),
                ("WARNING", 'All predictions are Correct prediction. To build a proper ErrorAnalyzer decision tree we need both correct and incorrect predictions'))
        np.testing.assert_array_equal(error_y, [
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.CORRECT_PREDICTION
        ])
        np.testing.assert_array_equal(error_rate, 0)



# -*- coding: utf-8 -*-
from mealy.constants import ErrorAnalyzerConstants
from sklearn.metrics import accuracy_score, balanced_accuracy_score
import numpy as np


def compute_confidence_decision(primary_model_true_accuracy, primary_model_predicted_accuracy):
    difference_true_pred_accuracy = np.abs(primary_model_true_accuracy - primary_model_predicted_accuracy)
    decision = difference_true_pred_accuracy <= ErrorAnalyzerConstants.TREE_ACCURACY_TOLERANCE

    fidelity = 1. - difference_true_pred_accuracy

    # TODO Binomial test
    return fidelity, decision


def compute_accuracy_score(y_true, y_pred):
    return accuracy_score(y_true, y_pred)


def compute_primary_model_accuracy(y):
    n_test_samples = y.shape[0]
    return float(np.count_nonzero(y == ErrorAnalyzerConstants.CORRECT_PREDICTION)) / n_test_samples


def compute_fidelity_score(y_true, y_pred):
    difference_true_pred_accuracy = np.abs(compute_primary_model_accuracy(y_true) -
                                           compute_primary_model_accuracy(y_pred))
    fidelity = 1. - difference_true_pred_accuracy

    return fidelity


def fidelity_balanced_accuracy_score(y_true, y_pred):
    return compute_fidelity_score(y_true, y_pred) + balanced_accuracy_score(y_true, y_pred)


def error_decision_tree_report(y_true, y_pred, output_format='str'):
    """Return a report showing the main Error Decision Tree metrics.

    Args:
        y_true (numpy.ndarray): Ground truth values of wrong/correct predictions of the error tree primary model.
            Expected values in [ErrorAnalyzerConstants.WRONG_PREDICTION, ErrorAnalyzerConstants.CORRECT_PREDICTION].
        y_pred (numpy.ndarray): Estimated targets as returned by the error tree. Expected values in
            [ErrorAnalyzerConstants.WRONG_PREDICTION, ErrorAnalyzerConstants.CORRECT_PREDICTION].
        output_format (string): Return format used for the report. Valid values are 'dict' or 'str'.

    Return:
        dict or str: dictionary or string report storing different metrics regarding the Error Decision Tree.
    """

    tree_accuracy_score = compute_accuracy_score(y_true, y_pred)
    tree_balanced_accuracy = balanced_accuracy_score(y_true, y_pred)
    primary_model_predicted_accuracy = compute_primary_model_accuracy(y_pred)
    primary_model_true_accuracy = compute_primary_model_accuracy(y_true)
    fidelity, confidence_decision = compute_confidence_decision(primary_model_true_accuracy,
                                                                primary_model_predicted_accuracy)
    if output_format == 'dict':
        report_dict = dict()
        report_dict[ErrorAnalyzerConstants.TREE_ACCURACY] = tree_accuracy_score
        report_dict[ErrorAnalyzerConstants.TREE_BALANCED_ACCURACY] = tree_balanced_accuracy
        report_dict[ErrorAnalyzerConstants.TREE_FIDELITY] = fidelity
        report_dict[ErrorAnalyzerConstants.PRIMARY_MODEL_TRUE_ACCURACY] = primary_model_true_accuracy
        report_dict[ErrorAnalyzerConstants.PRIMARY_MODEL_PREDICTED_ACCURACY] = primary_model_predicted_accuracy
        report_dict[ErrorAnalyzerConstants.CONFIDENCE_DECISION] = confidence_decision
        return report_dict

    if output_format == 'str':

        report = 'The Error Decision Tree was trained with accuracy %.2f%% and balanced accuracy %.2f%%.' % (tree_accuracy_score * 100, tree_balanced_accuracy * 100)
        report += '\n'
        report += 'The Decision Tree estimated the primary model''s accuracy to %.2f%%.' % \
                  (primary_model_predicted_accuracy * 100)
        report += '\n'
        report += 'The true accuracy of the primary model is %.2f.%%' % (primary_model_true_accuracy * 100)
        report += '\n'
        report += 'The Fidelity of the error tree is %.2f%%.' % \
                  (fidelity * 100)
        report += '\n'
        if not confidence_decision:
            report += 'Warning: the built tree might not be representative of the primary model performances.'
            report += '\n'
            report += 'The error tree predicted model accuracy is considered too different from the true model accuracy.'
            report += '\n'
        else:
            report += 'The error tree is considered representative of the primary model performances.'
            report += '\n'

        return report

    else:
        raise ValueError("Output format should either be 'dict' or 'str'")



from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder, StandardScaler, PowerTransformer, QuantileTransformer, \
    MaxAbsScaler, Binarizer, Normalizer, MinMaxScaler, RobustScaler


class ErrorAnalyzerConstants(object):

    WRONG_PREDICTION = "Wrong prediction"
    CORRECT_PREDICTION = "Correct prediction"
    MAX_DEPTH = [5, 10]
    MIN_SAMPLES_LEAF_LOWEST_UPPER_BOUND = .01 # for min_samples_leaf, the min upper bound value should be 0.01
    TEST_SIZE = 0.2

    MIN_NUM_ROWS = 100  # heuristic choice

    TREE_ACCURACY_TOLERANCE = 0.1
    CRITERION = 'entropy'
    NUMBER_EPSILON_VALUES = 50

    ERROR_TREE_COLORS = {CORRECT_PREDICTION: '#DDDDDD', WRONG_PREDICTION: '#CE1228'}

    TOP_K_FEATURES = 5

    TREE_ACCURACY = 'error_tree_accuracy_score'
    TREE_FIDELITY = 'error_tree_fidelity_score'
    TREE_BALANCED_ACCURACY = 'error_tree_balanced_accuracy_score'
    PRIMARY_MODEL_TRUE_ACCURACY = 'primary_model_true_accuracy'
    PRIMARY_MODEL_PREDICTED_ACCURACY = 'primary_model_predicted_accuracy'
    CONFIDENCE_DECISION = 'confidence_decision'

    NUMBER_PURITY_LEVELS = 10

    # use tuple because isinstance() takes only tuple as input type
    VALID_CATEGORICAL_STEPS = (OneHotEncoder, OrdinalEncoder)
    STEPS_THAT_DOES_NOT_CHANGE_OUTPUT_DIMENSION = (StandardScaler, PowerTransformer, QuantileTransformer, MaxAbsScaler,
                                                   Binarizer, Normalizer, MinMaxScaler, RobustScaler, SimpleImputer,
                                                   OrdinalEncoder)
    STEPS_THAT_CHANGE_OUTPUT_DIMENSION_WITH_OUTPUT_FEATURE_NAMES = (OneHotEncoder,)
    SUPPORTED_STEPS = STEPS_THAT_DOES_NOT_CHANGE_OUTPUT_DIMENSION + STEPS_THAT_CHANGE_OUTPUT_DIMENSION_WITH_OUTPUT_FEATURE_NAMES
    GRAPH_MAX_EDGE_WIDTH = 10
    GRAPH_MIN_LOCAL_ERROR_OPAQUE = 0.5
    
    # for imputers we don't need inverse function
    STEPS_THAT_CAN_BE_INVERSED_WITH_IDENTICAL_FUNCTION = (SimpleImputer,)



from unittest import TestCase
from unittest.mock import patch, Mock
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder
from .. import generate_preprocessing_steps


class TestFeatureTransformer(TestCase):
    def test_generate_preprocessing_steps_pipeline(self):
        ohe = Mock(spec=OneHotEncoder)
        pipe = Mock(spec=Pipeline, steps=[
            ("d", "drop"),
            ("p", "passthrough"),
            ("o", ohe)
        ])
        generator = generate_preprocessing_steps(pipe)
        self.assertEqual(next(generator), "passthrough")
        self.assertEqual(next(generator), ohe)
        with self.assertRaises(StopIteration):
            next(generator)

    def test_generate_preprocessing_steps_pipeline_inverted(self):
        ohe = Mock(spec=OneHotEncoder)
        pipe = Mock(spec=Pipeline, steps=[
            ("d", "drop"),
            ("p", "passthrough"),
            ("o", ohe)
        ])
        generator = generate_preprocessing_steps(pipe, True)
        self.assertEqual(next(generator), ohe)
        self.assertEqual(next(generator), "passthrough")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_generate_preprocessing_steps_only_drop(self):
        generator = generate_preprocessing_steps("drop")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_generate_preprocessing_steps_only_supported(self):
        ohe = Mock(spec=OneHotEncoder)
        generator = generate_preprocessing_steps(ohe)
        self.assertEqual(next(generator), ohe)
        with self.assertRaises(StopIteration):
            next(generator)

    def test_generate_preprocessing_steps_only_passthrough(self):
        generator = generate_preprocessing_steps("passthrough")
        self.assertEqual(next(generator), "passthrough")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_generate_preprocessing_steps_only_unsupported(self):
        msg = 'Mealy package does not support {}. '.format(Mock) + \
            'It might be because it changes output dimension without ' +\
            'providing a get_feature_names function to keep track of the ' + \
            'generated features, or that it does not provide an inverse_tranform method.'
        g = generate_preprocessing_steps(Mock())
        with self.assertRaisesRegex(TypeError, msg):
            next(g)



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, issparse
from unittest import TestCase
from unittest.mock import patch, Mock
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from sklearn.impute import SimpleImputer

from .. import PipelinePreprocessor, DummyPipelinePreprocessor


class TestFeatureTransformer(TestCase):
    def setUp(self):
        self.feature_list = ["num_1", "num_2", "cat_1", "cat_2"]


class TestDummyPipeline(TestFeatureTransformer):
    def setUp(self):
        super(TestDummyPipeline, self).setUp()
        self.x = np.array([[1, 2, 3], [2, 4, 6], [3, 6, 9]])
        self.df = pd.DataFrame(self.x, columns=self.feature_list[:-1])
        self.sparse_x = csr_matrix(self.x)
        self.pipe = DummyPipelinePreprocessor(self.feature_list)

    def test_transform_array(self):
        preprocessed_x = self.pipe.transform(self.x)
        np.testing.assert_array_equal(preprocessed_x, self.x)

    def test_transform_df(self):
        preprocessed_x = self.pipe.transform(self.df)
        np.testing.assert_array_equal(preprocessed_x, self.x)

    def test_transform_sparse(self):
        preprocessed_x = self.pipe.transform(self.sparse_x)
        self.assertEqual((preprocessed_x != self.sparse_x).nnz, 0)

    def test_get_top_ranked_feature_ids(self):
        input_to_results = {
            0: np.array([0, 2, 1]),
            2: np.array([0, 2]),
            -2: np.array([0])
        }
        for max_nr_features, result in input_to_results.items():
            with self.subTest(max_nr_features=max_nr_features):
                ranked = self.pipe.get_top_ranked_feature_ids(np.array([0.5, 0.2, 0.3]),
                                                              max_nr_features=max_nr_features)
                np.testing.assert_array_equal(ranked, result)

    def test_original_features(self):
        self.assertListEqual(self.pipe.get_original_feature_names(), self.feature_list)

    def test_preprocessed_features(self):
        self.assertListEqual(self.pipe.get_preprocessed_feature_names(), self.feature_list)

    def test_is_cat(self):
        self.assertFalse(self.pipe.is_categorical(0))

    def test_inverse_transform(self):
        array = np.random.rand(3,4)
        np.testing.assert_array_equal(self.pipe.inverse_transform(array), array)

    def test_inverse_transform_feature_id(self):
        self.assertEqual(self.pipe.inverse_transform_feature_id(1), 1)


class TestPreprocessingPipeline(TestFeatureTransformer):
    def setUp(self):
        super(TestPreprocessingPipeline, self).setUp()
        col_transformer = Mock(spec=ColumnTransformer)
        with patch("mealy.preprocessing.PipelinePreprocessor._get_feature_list_from_column_transformer", return_value=self.feature_list),\
            patch("mealy.preprocessing.PipelinePreprocessor._create_feature_mapping", return_value=None):
            self.pipe = PipelinePreprocessor(col_transformer)

    def test_init(self):
        col_transformer = Mock(spec=ColumnTransformer)
        with patch("mealy.preprocessing.PipelinePreprocessor._get_feature_list_from_column_transformer", return_value=self.feature_list),\
            patch("mealy.preprocessing.PipelinePreprocessor._create_feature_mapping", return_value=None):
            pipe_with_feature_arg = PipelinePreprocessor(col_transformer, self.feature_list)
            pipe_without_feature_arg = PipelinePreprocessor(col_transformer)
            with self.assertRaisesRegex(ValueError,
                "The list of features given by user does not correspond to the list of features handled by the Pipeline."):
                error = PipelinePreprocessor(col_transformer, ["cat_1", "num_2"])
        self.assertListEqual(pipe_with_feature_arg.get_original_feature_names(), self.feature_list)
        self.assertListEqual(pipe_without_feature_arg.get_original_feature_names(), self.feature_list)

    @patch("mealy.preprocessing.PipelinePreprocessor._update_feature_mapping_dict_using_input_names", return_value=None)
    @patch("mealy.preprocessing.PipelinePreprocessor._update_feature_mapping_dict_using_output_names", return_value=None)
    @patch("mealy.preprocessing.generate_preprocessing_steps", side_effect=lambda transformer: transformer)
    def test_create_feature_mapping_output_dim_change(self, _, mocked_o, mocked_i):
        ohe = Mock(spec=OneHotEncoder)
        steps = [
            Mock(spec=StandardScaler),
            "drop",
            "passthrough",
            ohe
        ]
        self.pipe.ct_preprocessor.transformers_ = [("does_not_matter", steps, ["num_1", "cat_2"])]
        self.pipe._create_feature_mapping()
        step, features, feat_ids = mocked_o.call_args[0]
        self.assertEqual(step, ohe)
        self.assertListEqual(features, ["num_1", "cat_2"])
        np.testing.assert_array_equal(feat_ids, np.array([0,3]))
        self.assertEqual(mocked_o.call_count, 1)
        self.assertEqual(mocked_i.call_count, 0)

    @patch("mealy.preprocessing.PipelinePreprocessor._update_feature_mapping_dict_using_input_names", return_value=None)
    @patch("mealy.preprocessing.PipelinePreprocessor._update_feature_mapping_dict_using_output_names", return_value=None)
    @patch("mealy.preprocessing.generate_preprocessing_steps", side_effect=lambda transformer: transformer)
    def test_create_feature_mapping_no_change_in_output_dim(self, _, mocked_o, mocked_i):
        steps = [
            Mock(spec=StandardScaler),
            "drop",
            "passthrough",
            Mock(spec=SimpleImputer)
        ]
        self.pipe.ct_preprocessor.transformers_ = [("does_not_matter", steps, self.feature_list)]
        self.pipe._create_feature_mapping()
        self.assertEqual(mocked_o.call_count, 0)
        self.assertEqual(mocked_i.call_count, 1)
        features, feat_ids = mocked_i.call_args[0]
        self.assertListEqual(features, self.feature_list)
        np.testing.assert_array_equal(feat_ids, np.array([0,1,2,3]))

    def test_update_feature_mapping_dict_using_input_names(self):
        self.pipe._update_feature_mapping_dict_using_input_names(["transformed_name_1", "transformed_name_2"], [0, 4])
        self.assertListEqual(self.pipe.preprocessed_feature_names, ["transformed_name_1", "transformed_name_2"])
        self.assertDictEqual(self.pipe.original2preprocessed, {0: [0], 4: [1]})
        self.assertDictEqual(self.pipe.preprocessed2original, {0: 0, 1: 4})

    def test_update_feature_mapping_dict_using_output_names(self):
        single_tr = Mock()
        single_tr.get_feature_names.return_value = ["very_transformed_name_1_0", "even_more_transformed_name"]
        self.pipe._update_feature_mapping_dict_using_output_names(single_tr, ["transformed_name_1", "transformed_name_2"], [0, 4])
        self.assertListEqual(self.pipe.preprocessed_feature_names, ["very_transformed_name_1_0", "even_more_transformed_name"])
        self.assertDictEqual(self.pipe.original2preprocessed, {0: [0]})
        self.assertDictEqual(self.pipe.preprocessed2original, {0: 0})

    def test_original_features(self):
        self.assertListEqual(self.pipe.get_original_feature_names(), self.feature_list)

    def test_is_cat(self):
        self.pipe.categorical_features = ["cat_1", "cat_2"]
        self.assertFalse(self.pipe.is_categorical(1))
        self.assertFalse(self.pipe.is_categorical(name="num_1"))
        self.assertTrue(self.pipe.is_categorical(3))
        self.assertTrue(self.pipe.is_categorical(name="cat_1"))
        with self.assertRaisesRegex(ValueError, "Either the input index or its name should be specified."):
            self.pipe.is_categorical()

    @patch("mealy.preprocessing.PipelinePreprocessor.inverse_transform_feature_id", side_effect=lambda idx: idx)
    @patch("mealy.preprocessing.PipelinePreprocessor.inverse_transform", side_effect=lambda a: a+1)
    def test_inverse_thresholds(self, mocked_inverse_transform, mocked_inverse_transform_feature_id):
        tree = Mock(feature=np.array([0,-2,1,3,-2,-2,0]), threshold=np.array([1, -2, 42,6,-2,-2,12]))
        self.pipe.preprocessed_feature_names = ["a", "b", "c", "d", "e", "f"]
        thresholds = self.pipe.inverse_thresholds(tree)
        a = mocked_inverse_transform.call_args[0][0]
        self.assertTrue(mocked_inverse_transform.call_count == 1)
        np.testing.assert_array_equal(a, np.array([
            [1,0,0,0,0,0],
            [0,42,0,0,0,0],
            [0,0,0,6,0,0],
            [12,0,0,0,0,0],
        ]))
        self.assertTrue(mocked_inverse_transform_feature_id.call_count == 4)
        first, second, third, fourth = mocked_inverse_transform_feature_id.call_args_list
        self.assertEqual(first[0][0], 0)
        self.assertEqual(second[0][0], 1)
        self.assertEqual(third[0][0], 3)
        self.assertEqual(fourth[0][0], 0)
        np.testing.assert_array_equal(thresholds, [2, -2, 43, 7, -2, -2, 13])

    def test_get_feature_ids_related_to_transformer(self):
        self.pipe.original2preprocessed = {0: [42], 1: [6], 2: [7], 3: [6, 7]}
        orig_feature_ids, preprocessed_feature_ids = self.pipe._get_feature_ids_related_to_transformer(["num_1", "cat_2"])
        self.assertListEqual(preprocessed_feature_ids, [42, 6, 7])
        np.testing.assert_array_equal(orig_feature_ids, [0, 3])

    def test_get_top_ranked_feature_ids(self):
        importance = np.array([20, 1, -2, 43])
        self.pipe.preprocessed2original = {0:1, 1:2, 2:3, 3:0}
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, 0), [0, 1, 2, 3])
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, -2), [0, 1])
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, 1), [0])

    def test_get_top_ranked_feature_ids_with_duplicates(self):
        importance = np.array([20, 1, -2, 43])
        self.pipe.preprocessed2original = {0:1, 1:1, 2:3, 3:3}
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, 0), [3, 1])
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, -1), [3, 1])
        self.assertListEqual(self.pipe.get_top_ranked_feature_ids(importance, 1), [3])

    @patch("mealy.preprocessing.generate_preprocessing_steps", side_effect=lambda transformer: transformer)
    def test_get_feature_list_from_column_transformer(self, _):
        steps = [
            "drop",
            "passthrough",
            Mock(spec=StandardScaler),
            Mock(spec=OneHotEncoder)
        ]

        other_steps = [
            "drop",
            "passthrough",
            Mock(spec=StandardScaler),
            Mock(spec=SimpleImputer)
        ]

        self.pipe.ct_preprocessor.transformers_ = [
            ("does_not_matter", steps, ["cat_1", "cat_2"]),
            ("does_not_matter", other_steps, ["num_1"])
        ]

        features = self.pipe._get_feature_list_from_column_transformer()
        self.assertListEqual(self.pipe.categorical_features, ["cat_1", "cat_2"])
        self.assertListEqual(features, ["cat_1", "cat_2", "num_1"])

    def test_inverse_transform(self):
        stsd = Mock(spec=StandardScaler)
        ohe = Mock(spec=OneHotEncoder)
        si = Mock(spec=SimpleImputer)

        self.pipe.ct_preprocessor.transformers_ = [
            ("does_not_matter", [stsd, ohe], ["num_1", "num_2"]),
            ("does_not_matter", [si], ["cat_1"])
        ]

        preprocessed_x = np.array([
            [0,1,2,3,4,5],
            [0,1,2,3,4,5],
            [0,1,2,3,4,5],
            [0,1,2,3,4,5],
            [0,1,2,3,4,5]
        ])

        expected = np.array([
            [5,6,2,0],
            [5,6,2,0],
            [5,6,2,0],
            [5,6,2,0],
            [5,6,2,0],
        ])

        def f(feature_names):
            if feature_names == ["cat_1"]:
                return 2, 1
            if feature_names == ["num_1", "num_2"]:
                return [0,1], [4,5]

        def g(step, transformer_output, feature_names):
            if step == si:
                return transformer_output * 2
            if step == ohe:
                return transformer_output + 4
            if step == stsd:
                return transformer_output - 3

        with patch("mealy.preprocessing.PipelinePreprocessor._get_feature_ids_related_to_transformer", side_effect=f),\
            patch("mealy.preprocessing.generate_preprocessing_steps", side_effect=lambda transformer, invert_order: reversed(transformer) if invert_order else transformer),\
            patch("mealy.preprocessing.PipelinePreprocessor._inverse_single_step", side_effect=g):
            result = self.pipe.inverse_transform(preprocessed_x)
        np.testing.assert_array_equal(result, expected)




import numpy as np
from unittest import TestCase, skip
from unittest.mock import Mock, patch

from .. import ErrorVisualizer, ErrorAnalyzer


class TestVisualizer(TestCase):
    def setUp(self):
        self.error_analyzer = Mock(spec=ErrorAnalyzer, _error_train_x=None, pipeline_preprocessor=Mock())


class TestPlotTree(TestVisualizer):
    def setUp(self):
        super(TestPlotTree, self).setUp()
        self.error_analyzer.error_tree.n_total_errors = 80
        self.error_analyzer.error_tree.error_class_idx = 1
        self.error_analyzer.error_tree.estimator_.tree_.value = np.array([
            [[40, 60]],
            [[5, 5]],
            [[0, 0]],
            [[0, 0]],
            [[30, 42]]
        ])
        self.error_analyzer.error_tree.estimator_.tree_.n_node_samples = np.array([100, 25, 40, 35, 72])
        self.error_analyzer.error_tree.estimator_.tree_.children_left = np.array([1, -2, -2, -2, -2])
        self.error_analyzer.error_tree.estimator_.tree_.children_right = np.array([4, -2, -2, -2, -2])
        self.error_analyzer.pipeline_preprocessor.get_original_feature_names.return_value = []

    def test_plot_error_tree(self):
        visualizer = ErrorVisualizer(self.error_analyzer)

        with patch("mealy.ErrorVisualizer.node_decision_rule",
            side_effect=lambda x, y: "decision rule" if y\
            else "pretty really very extremely fairly long decision rule"):
            result = visualizer.plot_error_tree((20, 30)).source.split("\n")

        digraph_desc, size, node_desc, edge_desc, graph_desc, leave_desc, final_line = result[:5] + result[-2:]
        root_label, root_samples, root_local_error, root_global_error, root_alpha = result[5:10]
        left_label, left_rule, left_samples, left_local_error, left_global_error, left_alpha, left_edge = result[10:17]
        right_label, right_rule, right_samples, right_local_error, right_global_error, right_alpha, right_edge = result[17:24]

        assert digraph_desc == 'digraph Tree {'
        assert size == ' size="20,30!";'
        assert node_desc == 'node [shape=box, style="filled, rounded", color="black", fontname=helvetica] ;'
        assert edge_desc == 'edge [fontname=helvetica] ;'
        assert graph_desc == 'graph [ranksep=equally, splines=polyline] ;'
        assert leave_desc == '{rank=same ; 1; 4} ;'
        assert final_line == '}'

        assert root_label == '0 [label="node #0'
        assert root_samples == 'samples = 100%'
        assert root_local_error == 'local error = 60%'
        assert root_global_error == 'fraction of total error = 75%'
        assert root_alpha == '", fillcolor="#CE122899", tooltip="root"] ;'

        assert left_label == '1 [label="node #1'
        assert left_rule == 'decision rule'
        assert left_samples == 'samples = 25%'
        assert left_local_error == 'local error = 20%'
        assert left_global_error == 'fraction of total error = 6.25%'
        assert left_alpha == '", fillcolor="#CE122833", tooltip="decision rule"] ;'
        assert left_edge == '0 -> 1 [penwidth=1];'

        assert right_label == '4 [label="node #4'
        assert right_rule == 'pretty really very extremely fai...'
        assert right_samples == 'samples = 72%'
        assert right_local_error == 'local error = 58.333%'
        assert right_global_error == 'fraction of total error = 52.5%'
        assert right_alpha == '", fillcolor="#CE122894", tooltip="pretty really very extremely fairly long decision rule"] ;'
        assert right_edge == '0 -> 4 [penwidth=5.25];'


class TestNodeDecisionRule(TestVisualizer):
    def setUp(self):
        super(TestNodeDecisionRule, self).setUp()
        self.error_analyzer.pipeline_preprocessor.get_original_feature_names.return_value = ["num", "cat"]
        self.error_analyzer.pipeline_preprocessor.is_categorical.side_effect = lambda name: name=="cat"
        self.error_analyzer._inverse_transform_thresholds.return_value = ["A", 6.45678]
        self.error_analyzer._inverse_transform_features.return_value = [1, 0]
        self.visualizer = ErrorVisualizer(self.error_analyzer)

    def test_categorical_left_node(self):
        cat_left = self.visualizer.node_decision_rule(0, True)
        self.assertEqual(cat_left, "cat is A")

    def test_categorical_right_node(self):
        cat_right = self.visualizer.node_decision_rule(0, False)
        self.assertEqual(cat_right, "cat is not A")

    def test_numerical_left_node(self):
        num_left = self.visualizer.node_decision_rule(1, True)
        self.assertEqual(num_left, "num <= 6.46")

    def test_numerical_right_node(self):
        num_right = self.visualizer.node_decision_rule(1, False)
        self.assertEqual(num_right, "6.46 < num")


# TODO: finish the tests
@skip(reason="Not done yet")
class TestPlotDistribution(TestVisualizer):
    def setUp(self):
        super(TestPlotDistribution, self).setUp()

    def test_show_global_show_class(self):
        pass

    def test_show_global_no_show_class(self):
        pass

    def test_no_show_global_show_class(self):
        pass

    def test_no_show_global_no_show_class(self):
        pass



import numpy as np
from unittest import TestCase
from unittest.mock import Mock

from sklearn.exceptions import NotFittedError
from .. import ErrorTree, ErrorAnalyzerConstants


class TestErrorTree(TestCase):
    def test_empty_tree(self):
        with self.assertRaisesRegex(NotFittedError, "You should fit the ErrorAnalyzer first"):
            ErrorTree(None)

    def test_small_tree(self):
        clf = Mock()
        clf.tree_.node_count = 1
        with self.assertLogs("mealy.error_tree", level="WARNING") as caplog:
            error_tree = ErrorTree(clf)
            self.assertEqual(caplog.output, [
                "WARNING:mealy.error_tree:The error tree has only one node, " +
                "there will be problems when using it with ErrorVisualizer"
            ])

    def test_tree(self):
        tree_ = Mock(value=np.array([
            [[42, 69]],
            [[2, 9]],
            [[40, 58]],
            [[30, 0]],
            [[10, 10]],
            [[2, 9]],
            [[8, 1]],
            [[2, 30]],
            [[28, 28]],
            [[1, 10]],
            [[27, 18]]
            ]), feature=np.array([1, 2, -2, -2, 0, -2, -2, 0, 1]))

        clf = Mock(classes_=np.array([ErrorAnalyzerConstants.CORRECT_PREDICTION,
            ErrorAnalyzerConstants.WRONG_PREDICTION]), tree_=tree_)

        error_tree = ErrorTree(clf)

        self.assertEqual(error_tree.error_class_idx, 1)
        self.assertEqual(error_tree.n_total_errors, 69)
        np.testing.assert_array_equal(error_tree.leaf_ids, [2, 3, 5, 6])

        # Ranking arrays
        np.testing.assert_array_equal(error_tree.correctly_predicted_leaves, [40, 30, 2, 8])
        np.testing.assert_array_equal(error_tree.wrongly_predicted_leaves, [58, 0, 9, 1])
        np.testing.assert_array_equal(error_tree.difference, [-18, 30, -7, 7])
        np.testing.assert_array_equal(error_tree.impurity, [20/49, 1, 2/11, 8/9])
        np.testing.assert_array_equal(error_tree.quantized_impurity, [4, 10, 2, 9])
        np.testing.assert_array_equal(error_tree.total_error_fraction, [58/69, 0, 9/69, 1/69])



# -*- coding: utf-8 -*-
import numpy as np
import collections
from sklearn import tree
from sklearn.model_selection import GridSearchCV
from sklearn.base import is_regressor
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.base import BaseEstimator
from sklearn.metrics import make_scorer
from .error_analysis_utils import check_enough_data, get_epsilon, format_float
from .constants import ErrorAnalyzerConstants
from .metrics import error_decision_tree_report, fidelity_balanced_accuracy_score
from .preprocessing import PipelinePreprocessor, DummyPipelinePreprocessor
from .error_tree import ErrorTree
from sklearn.exceptions import NotFittedError

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='mealy | %(levelname)s - %(message)s')


class ErrorAnalyzer(BaseEstimator):
    """ ErrorAnalyzer analyzes the errors of a prediction model on a test set.

    It uses model predictions and ground truth target to compute the model errors on the test set.
    It then trains a Decision Tree, called a Error Analyzer Tree, on the same test set by using the model error
    as target. The nodes of the decision tree are different segments of errors to be studied individually.

    Args:
        primary_model (sklearn.base.BaseEstimator or sklearn.pipeline.Pipeline): a sklearn model to analyze. Either an estimator
            or a Pipeline containing a ColumnTransformer with the preprocessing steps and an estimator as last step.
        feature_names (list of str): list of feature names. Defaults to None.
        param_grid (dict): sklearn.tree.DecisionTree hyper-parameters values for grid search.
        random_state (int): random seed.

    Attributes:
        _error_tree (DecisionTreeClassifier): the estimator used to train the Error Analyzer Tree
    """

    def __init__(self, primary_model,
                feature_names=None,
                param_grid=None,
                probability_threshold=0.5,
                random_state=65537):
        self.param_grid = param_grid
        self.probability_threshold = probability_threshold
        self.random_state = random_state

        if isinstance(primary_model, Pipeline):
            if len(primary_model.steps) != 2:
                logger.warning("Pipeline should have two steps: the preprocessing of the features, and the primary model to analyze.")
            estimator = primary_model.steps[-1][1]
            if not isinstance(estimator, BaseEstimator):
                raise TypeError("The last step of the pipeline has to be a BaseEstimator.")
            self._primary_model = estimator

            ct_preprocessor = primary_model.steps[0][1]
            if not isinstance(ct_preprocessor, ColumnTransformer):
                raise TypeError("The input preprocessor has to be a ColumnTransformer.")
            self.pipeline_preprocessor = PipelinePreprocessor(ct_preprocessor, feature_names)
        elif isinstance(primary_model, BaseEstimator):
            self._primary_model = primary_model
            self.pipeline_preprocessor = DummyPipelinePreprocessor(feature_names)
        else:
            raise TypeError('ErrorAnalyzer needs as input either a scikit BaseEstimator or a scikit Pipeline.')

        self._error_tree = None
        self._error_train_x = None
        self._error_train_y = None
        self.epsilon = None

    @property
    def param_grid(self):
        return self._param_grid

    @param_grid.setter
    def param_grid(self, value):
        self._param_grid = value

    @property
    def random_state(self):
        return self._random_state

    @random_state.setter
    def random_state(self, value):
        self._random_state = value

    @property
    def error_tree(self):
        if self._error_tree is None:
            raise NotFittedError("The error tree is not fitted yet. Call 'fit' method with appropriate arguments before using this estimator.")
        return self._error_tree

    @error_tree.setter
    def error_tree(self, tree):
        if self.pipeline_preprocessor.get_preprocessed_feature_names() is None:
            self.pipeline_preprocessor.preprocessed_feature_names = ["feature#%s" % feature_index
                for feature_index in range(tree.estimator_.n_features_)]
        self._error_tree = tree

    @property
    def preprocessed_feature_names(self):
        return self.pipeline_preprocessor.get_preprocessed_feature_names()

    def fit(self, X, y):
        """
        Fit the Error Analyzer Tree.

        Trains the Error Analyzer Tree, a Decision Tree to discriminate between samples that are correctly
        predicted or wrongly predicted (errors) by a primary model.

        Args:
            X (numpy.ndarray or pandas.DataFrame): feature data from a test set to evaluate the primary predictor and
                train a Error Analyzer Tree.
            y (numpy.ndarray or pandas.DataFrame): target data from a test set to evaluate the primary predictor and
                train a Error Analyzer Tree.
        """
        logger.info("Preparing the Error Analyzer Tree...")

        np.random.seed(self._random_state)
        preprocessed_X = self.pipeline_preprocessor.transform(X)

        check_enough_data(preprocessed_X, min_len=ErrorAnalyzerConstants.MIN_NUM_ROWS)
        self._error_train_y, error_rate = self._compute_primary_model_error(preprocessed_X, y)
        self._error_train_x = preprocessed_X

        logger.info("Fitting the Error Analyzer Tree...")
        # entropy/mutual information is used to split nodes in Microsoft Pandora system
        dt_clf = tree.DecisionTreeClassifier(criterion=ErrorAnalyzerConstants.CRITERION,
                                             random_state=self._random_state)
        param_grid = self.param_grid
        if param_grid is None:
            min_samples_leaf_max = min(error_rate, ErrorAnalyzerConstants.MIN_SAMPLES_LEAF_LOWEST_UPPER_BOUND)
            param_grid = {
                'max_depth': ErrorAnalyzerConstants.MAX_DEPTH,
                'min_samples_leaf': np.linspace(min_samples_leaf_max/5, min_samples_leaf_max, 5)
            }

        logger.info('Grid search the Error Tree with the following grid: {}'.format(param_grid))
        gs_clf = GridSearchCV(dt_clf,
                              param_grid=param_grid,
                              cv=5,
                              scoring=make_scorer(fidelity_balanced_accuracy_score))

        gs_clf.fit(self._error_train_x, self._error_train_y)
        self.error_tree = ErrorTree(error_decision_tree=gs_clf.best_estimator_)
        logger.info('Chosen parameters: {}'.format(gs_clf.best_params_))

    def get_error_leaf_summary(self, leaf_selector=None, add_path_to_leaves=False,
                               output_format='dict', rank_by='total_error_fraction'):
        """ Return summary information regarding leaves.

        Args:
            leaf_selector (None, int or array-like): The leaves whose information will be returned
                * int: Only return information of the leaf with the corresponding id
                * array-like: Only return information of the leaves corresponding to these ids
                * None (default): Return information of all the leaves
            add_path_to_leaves (bool): Whether to add information of the path across the tree till the selected node. Defaults to False.
            output_format (string): Return format used for the report. Valid values are 'dict' or 'str'. Defaults to 'dict'.
            rank_by (str): Ranking criterion for the leaves. Valid values are:
                * 'total_error_fraction' (default): rank by the fraction of total error in the node
                * 'purity': rank by the purity (ratio of wrongly predicted samples over the total number of node samples)
                * 'class_difference': rank by the difference of number of wrongly and correctly predicted samples
                in a node.

        Return:
            dict or str: list of reports (as dictionary or string) with different information on each selected leaf.
        """

        leaf_nodes = self._get_ranked_leaf_ids(leaf_selector=leaf_selector, rank_by=rank_by)

        leaves_summary = []
        for leaf_id in leaf_nodes:
            n_errors = int(self.error_tree.estimator_.tree_.value[leaf_id, 0, self.error_tree.error_class_idx])
            n_samples = self.error_tree.estimator_.tree_.n_node_samples[leaf_id]
            local_error = n_errors / n_samples
            total_error_fraction = n_errors / self.error_tree.n_total_errors
            n_corrects = n_samples - n_errors

            if output_format == 'dict':
                leaf_dict = {
                    "id": leaf_id,
                    "n_corrects": n_corrects,
                    "n_errors": n_errors,
                    "local_error": local_error,
                    "total_error_fraction": total_error_fraction
                }
                if add_path_to_leaves:
                    leaf_dict["path_to_leaf"] = self._get_path_to_node(leaf_id)
                leaves_summary.append(leaf_dict)

            elif output_format == 'str':
                leaf_summary = 'LEAF %d:\n' % leaf_id
                leaf_summary += '     Correct predictions: %d | Wrong predictions: %d | Local error (purity): %.2f | Fraction of total error: %.2f\n' % (n_corrects, n_errors, local_error, total_error_fraction)

                if add_path_to_leaves:
                    leaf_summary += '     Path to leaf:\n'
                    for (step_idx, step) in enumerate(self._get_path_to_node(leaf_id)):
                        leaf_summary += '     ' + '   ' * step_idx + step  + '\n'

                leaves_summary.append(leaf_summary)

            else:
                raise ValueError("Output format should either be 'dict' or 'str'")

        return leaves_summary

    def evaluate(self, X, y, output_format='str'):
        """
        Evaluate performance of ErrorAnalyzer on the given test data and labels.
        Return ErrorAnalyzer summary metrics regarding the Error Tree.

        Args:
            X (numpy.ndarray or pandas.DataFrame): feature data from a test set to evaluate the primary predictor
                and train a Error Analyzer Tree.
            y (numpy.ndarray or pandas.DataFrame): target data from a test set to evaluate the primary predictor and
                train a Error Analyzer Tree.
            output_format (string): Return format used for the report. Valid values are 'dict' or 'str'. Defaults to 'str'.

        Return:
            dict or str: dictionary or string report storing different metrics regarding the Error Decision Tree.
        """
        prep_x, prep_y = self.pipeline_preprocessor.transform(X), np.array(y)
        y_pred = self.error_tree.estimator_.predict(prep_x)
        y_true, _ = self._compute_primary_model_error(prep_x, prep_y)
        return error_decision_tree_report(y_true, y_pred, output_format)

    def _compute_primary_model_error(self, X, y):
        """
        Computes the errors of the primary model predictions and samples

        Args:
            X: array-like of shape (n_samples, n_features)
            Input samples.

            y: array-like of shape (n_samples,)
            True target values for `X`.

        Returns:
             sampled_X: ndarray
             A sample of `X`.

             error_y: array of string of shape (n_sampled_X, )
             Boolean value of whether or not the primary model predicted correctly or incorrectly the samples in sampled_X.
        """
        if is_regressor(self._primary_model) or len(np.unique(y)) > 2:
            # regression or multiclass classification models: no proba threshold
            y_pred = self._primary_model.predict(X)
        else: # binary -> need to check the proba threshold
            prediction_index = (self._primary_model.predict_proba(X)[:, 1] > self.probability_threshold).astype(int)
            # map the prediction indexes to the original target values
            y_pred = np.array([self._primary_model.classes_[i] for i in prediction_index])

        error_y, error_rate = self._evaluate_primary_model_predictions(y_true=y, y_pred=y_pred)
        return error_y, error_rate

    def _evaluate_primary_model_predictions(self, y_true, y_pred):
        """
        Compute errors of the primary model on the test set

        Args:
            y_true: 1D array
            True target values.

            y_pred: 1D array
            Predictions of the primary model.

        Return:
            error_y: array of string of len(y_true)
            Boolean value of whether or not the primary model got the prediction right.

            error_rate: float
            Accuracy of the primary model
        """

        error_y = np.full_like(y_true, ErrorAnalyzerConstants.CORRECT_PREDICTION, dtype="O")
        if is_regressor(self._primary_model):
            difference = np.abs(y_true - y_pred)
            if self.epsilon is None:
                # only compute epsilon when fitting the model (not while evaluating)
                self.epsilon = get_epsilon(difference)
            error_mask = difference > self.epsilon
        else:
            error_mask = y_true != y_pred

        n_wrong_preds = np.count_nonzero(error_mask)
        error_y[error_mask] = ErrorAnalyzerConstants.WRONG_PREDICTION

        if n_wrong_preds == 0 or n_wrong_preds == len(error_y):
            logger.warning('All predictions are {}. To build a proper ErrorAnalyzer decision tree we need both correct and incorrect predictions'.format(error_y[0]))

        error_rate = n_wrong_preds / len(error_y)
        logger.info('The primary model has an error rate of {}'.format(format_float(error_rate, 3)))
        return error_y, error_rate

    def _get_ranked_leaf_ids(self, leaf_selector=None, rank_by='total_error_fraction'):
        """ Select error nodes and rank them by importance.

        Args:
            leaf_selector (None, int or array-like): the leaves whose information will be returned
                * int: Only return information of the leaf with the corresponding id
                * array-like: Only return information of the leaves corresponding to these ids
                * None (default): Return information of all the leaves
            rank_by (str): ranking criterion for the leaves. Valid values are:
                * 'total_error_fraction': rank by the fraction of total error in the node
                * 'purity': rank by the purity (ratio of wrongly predicted samples over the total number of node samples)
                * 'class_difference': rank by the difference of number of wrongly and correctly predicted samples
                in a node.

        Return:
            list or numpy.ndarray: list of selected leaves indices.

        """
        apply_leaf_selector = self._get_leaf_selector(self.error_tree.leaf_ids, leaf_selector)
        selected_leaves = apply_leaf_selector(self.error_tree.leaf_ids)
        if selected_leaves.size == 0:
            return selected_leaves
        if rank_by == 'total_error_fraction':
            sorted_ids = np.argsort(-apply_leaf_selector(self.error_tree.total_error_fraction))
        elif rank_by == 'purity':
            sorted_ids = np.lexsort((apply_leaf_selector(self.error_tree.difference), apply_leaf_selector(self.error_tree.quantized_impurity)))
        elif rank_by == 'class_difference':
            sorted_ids = np.lexsort((apply_leaf_selector(self.error_tree.impurity), apply_leaf_selector(self.error_tree.difference)))
        else:
            raise ValueError(
                "Input argument for rank_by is invalid. Should be 'total_error_fraction', 'purity' or 'class_difference'")
        return selected_leaves.take(sorted_ids)

    @staticmethod
    def _get_leaf_selector(leaf_ids, leaf_selector=None):
        """
        Return a function that select rows of provided arrays. Arrays must be of shape (1, number of leaves)
            Args:
                leaf_selector: None, int or array-like
                    How to select the rows of the array
                      * int: Only keep the row corresponding to this leaf id
                      * array-like: Only keep the rows corresponding to these leaf ids
                      * None (default): Keep the whole array of leaf ids

            Return:
                A function with one argument array as a selector of leaf ids
                Args:
                    array: numpy array of shape (1, number of leaves)
                    An array of which we only want to keep some rows
        """
        if leaf_selector is None:
            return lambda array: array

        leaf_selector_as_array = np.array(leaf_selector)
        leaf_selector = np.in1d(leaf_ids, leaf_selector_as_array)
        nr_kept_leaves = np.count_nonzero(leaf_selector)
        if nr_kept_leaves == 0:
            logger.info("None of the ids provided correspond to a leaf id.")
        elif nr_kept_leaves < leaf_selector_as_array.size:
            logger.info("Some of the ids provided do not belong to leaves. Only leaf ids are kept.")
        return lambda array: array[leaf_selector]

    def _get_path_to_node(self, node_id):
        """ Return path to node as a list of split steps from the nodes of the sklearn Tree object """
        feature_names = self.pipeline_preprocessor.get_original_feature_names()
        children_left = list(self.error_tree.estimator_.tree_.children_left)
        children_right = list(self.error_tree.estimator_.tree_.children_right)
        threshold = self._inverse_transform_thresholds()
        feature = self._inverse_transform_features()

        cur_node_id = node_id
        path_to_node = collections.deque()
        while cur_node_id > 0:

            node_is_left_child = cur_node_id in children_left
            if node_is_left_child:
                parent_id = children_left.index(cur_node_id)
            else:
                parent_id = children_right.index(cur_node_id)

            feat = feature[parent_id]
            thresh = threshold[parent_id]

            is_categorical = self.pipeline_preprocessor.is_categorical(feat)
            thresh = str(thresh) if is_categorical else format_float(thresh, 2)

            decision_rule = ''
            if node_is_left_child:
                decision_rule += ' <= ' if not is_categorical else ' is not '
            else:
                decision_rule += " > " if not is_categorical else ' is '

            decision_rule = str(feature_names[feat]) + decision_rule + thresh
            path_to_node.appendleft(decision_rule)
            cur_node_id = parent_id

        return path_to_node

    def _inverse_transform_features(self):
        """ Undo preprocessing of feature values.

        If the predictor comes with a Pipeline preprocessor, map the features indices of the Error Analysis
        Tree back to their indices in the original unpreprocessed space of features.
        Otherwise simply return the feature indices of the decision tree. The feature indices of a decision tree
        indicate what features are used to split the training set at each node.
        See https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html.

        Return:
            list or numpy.ndarray:
                indices of features of the Error Analyzer Tree, possibly mapped back to the
                original unprocessed feature space.
        """
        return [self.pipeline_preprocessor.inverse_transform_feature_id(feat_idx) if feat_idx > 0 else feat_idx
            for feat_idx in self.error_tree.estimator_.tree_.feature]

    def _inverse_transform_thresholds(self):
        """  Undo preprocessing of feature threshold values.

        If the predictor comes with a Pipeline preprocessor, undo the preprocessing on the thresholds of the Error Analyzer
        Tree for an easier plot interpretation. Otherwise simply return the thresholds of
        the decision tree. The thresholds of a decision tree are the feature values used to split the training set at
        each node. See https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html.

        Return:
            numpy.ndarray:
                thresholds of the Error Tree, possibly with preprocessing undone.
        """
        return self.pipeline_preprocessor.inverse_thresholds(self.error_tree.estimator_.tree_)



# -*- coding: utf-8 -*-
from sklearn.pipeline import Pipeline
import numpy as np
from mealy.constants import ErrorAnalyzerConstants
from kneed import KneeLocator


def get_epsilon(difference):
    """
    Compute the threshold used to decide whether a prediction is wrong or correct (for regression tasks).

    Args:
           difference (1D-array): The absolute differences between the true target values and the predicted ones (by the primary model).

    Return:
           epsilon (float): The value of the threshold used to decide whether the prediction for a regression task is wrong or correct
    """
    epsilon_range = np.linspace(min(difference), max(difference), num=ErrorAnalyzerConstants.NUMBER_EPSILON_VALUES)
    cdf_error = []
    n_samples = difference.shape[0]
    for epsilon in epsilon_range:
        correct_predictions = difference <= epsilon
        cdf_error.append(np.count_nonzero(correct_predictions) / float(n_samples))
    return KneeLocator(epsilon_range, cdf_error).knee

def generate_preprocessing_steps(transformer, invert_order=False):
    if isinstance(transformer, Pipeline):
        steps = [step for name, step in transformer.steps]
        if invert_order:
            steps = reversed(steps)
    else:
        steps = [transformer]
    for step in steps:
        if step == 'drop':
            # Skip the drop step of ColumnTransformer
            continue
        if step != 'passthrough' and not isinstance(step, ErrorAnalyzerConstants.SUPPORTED_STEPS):
            # Check all the preprocessing steps are supported by mealy
            unsupported_class = step.__class__
            raise TypeError('Mealy package does not support {}. '.format(unsupported_class) +
                        'It might be because it changes output dimension without ' +
                        'providing a get_feature_names function to keep track of the ' +
                        'generated features, or that it does not provide an ' +
                        'inverse_tranform method.')
        yield step

def invert_transform_via_identity(step):
    if isinstance(step, ErrorAnalyzerConstants.STEPS_THAT_CAN_BE_INVERSED_WITH_IDENTICAL_FUNCTION):
        return True
    if step == 'passthrough' or step is None:
        return True
    return False

def check_lists_having_same_elements(list_A, list_B):
    return set(list_A) == set(list_B)

def check_enough_data(df, min_len):
    """
    Compare length of dataframe to minimum lenght of the test data.
    Used in the relevance of the measure.

    :param df: Input dataframe
    :param min_len:
    :return:
    """
    if df.shape[0] < min_len:
        raise ValueError(
            'The original dataset is too small ({} rows) to have stable result, it needs to have at least {} rows'.format(
                df.shape[0], min_len))

def format_float(number, decimals):
    """
    Format a number to have the required number of decimals. Ensure no trailing zeros remain.

    Args:
        number (float or int): The number to format
        decimals (int): The number of decimals required

    Return:
        formatted (str): The number as a formatted string

    """
    formatted = ("{:." + str(decimals) + "f}").format(number).rstrip("0")
    if formatted.endswith("."):
        return formatted[:-1]
    return formatted



# Dataiku Model Error Analysis lib
from .version import __version__
from .constants import ErrorAnalyzerConstants
from .error_analyzer import ErrorAnalyzer
from .error_visualizer import ErrorVisualizer, _BaseErrorVisualizer
from .error_tree import ErrorTree



# -*- coding: utf-8 -*-
import numpy as np
from graphviz import Source
from collections import deque
import matplotlib.pyplot as plt
from .constants import ErrorAnalyzerConstants
from .error_analyzer import ErrorAnalyzer
from .error_analysis_utils import format_float

plt.rc('font', family="sans-serif")
SMALL_SIZE, MEDIUM_SIZE, BIGGER_SIZE = 8, 10, 12
plt.rc('axes', titlesize=BIGGER_SIZE, labelsize=MEDIUM_SIZE)
plt.rc('xtick', labelsize=SMALL_SIZE)
plt.rc('ytick', labelsize=SMALL_SIZE)
plt.rc('legend', fontsize=SMALL_SIZE)
plt.rc("hatch", color="white", linewidth=4)


class _BaseErrorVisualizer(object):
    def __init__(self, error_analyzer):
        if not isinstance(error_analyzer, ErrorAnalyzer):
            raise TypeError('You need to input an ErrorAnalyzer object.')

        self._error_analyzer = error_analyzer

        self._get_ranked_leaf_ids = lambda leaf_selector, rank_by: \
            error_analyzer._get_ranked_leaf_ids(leaf_selector, rank_by)

    @staticmethod
    def _plot_histograms(hist_data, label, **params):
        bottom = None
        for class_value in [ErrorAnalyzerConstants.CORRECT_PREDICTION, ErrorAnalyzerConstants.WRONG_PREDICTION]:
            bar_heights = hist_data.get(class_value)
            if bar_heights is not None:
                plt.bar(height=bar_heights,
                        label="{} ({})".format(class_value, label),
                        edgecolor="white",
                        linewidth=1,
                        color=ErrorAnalyzerConstants.ERROR_TREE_COLORS[class_value],
                        bottom=bottom,
                        **params)
                bottom = bar_heights

    @staticmethod
    def _add_new_plot(figsize, bins, x_ticks, feature_name, suptitle):
        plt.figure(figsize=figsize)
        plt.xticks(x_ticks, rotation="45")
        plt.gca().set_xticklabels(labels=bins)
        plt.ylabel('Proportion of samples')
        plt.title('Distribution of {}'.format(feature_name))
        plt.suptitle(suptitle)

    @staticmethod
    def _plot_feature_distribution(x_ticks, feature_is_numerical, leaf_data, root_data=None):
        width, x = 1.0, x_ticks
        align = "edge"
        if root_data is not None:
            width /= 2
            if feature_is_numerical:
                x = x_ticks[1:]
            _BaseErrorVisualizer._plot_histograms(root_data, label="global data", x=x, hatch="///",
                                                  width=-width, align=align)
        if leaf_data is not None:
            if feature_is_numerical:
                x = x_ticks[:-1]
            elif root_data is None:
                align = "center"
            _BaseErrorVisualizer._plot_histograms(leaf_data, label="leaf data", x=x,
                                                  align=align, width=width)
        plt.legend()
        plt.pause(0.05)


class ErrorVisualizer(_BaseErrorVisualizer):
    """
    ErrorVisualizer provides visual utilities to analyze the Error Tree in ErrorAnalyzer

    Args:
        error_analyzer (ErrorAnalyzer): fitted ErrorAnalyzer representing the performance of a primary model.
    """

    def __init__(self, error_analyzer):
        super(ErrorVisualizer, self).__init__(error_analyzer)

        self._error_tree = self._error_analyzer.error_tree
        self._error_clf = self._error_tree.estimator_
        self._train_leaf_ids = self._error_clf.apply(self._error_analyzer._error_train_x)
        self._thresholds = None
        self._features = None

        self._original_feature_names = self._error_analyzer.pipeline_preprocessor.get_original_feature_names()
        self._numerical_feature_names = [f for f in self._original_feature_names if not self._error_analyzer.pipeline_preprocessor.is_categorical(name=f)]

    @property
    def thresholds_(self):
        if self._thresholds is None:
            self._thresholds = self._error_analyzer._inverse_transform_thresholds()
        return self._thresholds

    @property
    def features_(self):
        if self._features is None:
            self._features = self._error_analyzer._inverse_transform_features()
        return self._features

    def plot_error_tree(self, size=(50, 50)):
        """
        Plot the graph of the decision tree.

        Args:
            size (tuple): size of the output plot.

        Return:
            graphviz.Source: graph of the Error Analyzer Tree.

        """
        dot_str = 'digraph Tree {{\n size="{0},{1}!";\n'.format(size[0], size[1])
        dot_str += 'node [shape=box, style="filled, rounded", color="black", fontname=helvetica] ;\n'
        dot_str += 'edge [fontname=helvetica] ;\ngraph [ranksep=equally, splines=polyline] ;\n'
        color = ErrorAnalyzerConstants.ERROR_TREE_COLORS[ErrorAnalyzerConstants.WRONG_PREDICTION]
        leaves, left_child_to_parent, right_child_to_parent = set(), {}, {}
        ids = deque()
        ids.append(0)

        while ids:
            node_id = ids.popleft()
            dot_str += '{0} [label="node #{0}\n'.format(node_id)

            parent_id = left_child_to_parent.get(node_id, right_child_to_parent.get(node_id))
            if parent_id is not None:
                rule = self.node_decision_rule(parent_id, node_id in left_child_to_parent)
                dot_str += "{}\n".format((rule[:32] + "...") if len(rule) > 35 else rule)

            n_wrong_preds = self._error_clf.tree_.value[node_id, 0, self._error_tree.error_class_idx]
            total_error_fraction = n_wrong_preds / self._error_tree.n_total_errors
            samples = self._error_clf.tree_.n_node_samples[node_id]
            local_error = n_wrong_preds / samples
            dot_str += 'samples = {}%\n'.format(format_float(100 * samples / self._error_clf.tree_.n_node_samples[0], 3))
            dot_str += 'local error = {}%\n'.format(format_float(100 * local_error, 3))
            dot_str += 'fraction of total error = {}%\n'.format(format_float(100 * total_error_fraction, 3))

            alpha = "{:02x}".format(int(local_error*255))
            dot_str += '", fillcolor="{}", tooltip="{}"] ;\n'.format(color+alpha, "root" if parent_id is None else rule)

            if parent_id is not None:
                edge_width = max(1, ErrorAnalyzerConstants.GRAPH_MAX_EDGE_WIDTH * total_error_fraction)
                dot_str += '{} -> {} [penwidth={}];\n'.format(parent_id, node_id, edge_width)
            left_child_id, right_child_id = self._error_clf.tree_.children_left[node_id], self._error_clf.tree_.children_right[node_id]
            if left_child_id > 0:
                ids += [left_child_id, right_child_id]
                left_child_to_parent[left_child_id] = node_id
                right_child_to_parent[right_child_id] = node_id
            else:
                leaves.add(node_id)

        dot_str += '{rank=same ; '+ '; '.join(map(str, leaves)) + '} ;\n'
        dot_str += "}"
        return Source(dot_str)

    def node_decision_rule(self, parent_id, left_child):
        feature = self._original_feature_names[self.features_[parent_id]]
        value = self.thresholds_[parent_id]
        numerical_split = feature in self._numerical_feature_names
        if numerical_split:
            if left_child:
                return '{} <= {}'.format(feature, format_float(value, 2))
            return '{} < {}'.format(format_float(value, 2), feature)
        return feature + ' is ' + ( '' if left_child else 'not ') + str(value)

    def plot_feature_distributions_on_leaves(self, leaf_selector=None,
                                             top_k_features=ErrorAnalyzerConstants.TOP_K_FEATURES,
                                             show_global=True, show_class=True, rank_leaves_by="total_error_fraction",
                                             nr_bins=10, figsize=(15, 10)):

        """
        Return feature distribution plots at the selected leaves.

        The leaves for which the distributions are plotted are determined by the leaf_selector argument.
        By default, no specific leaves are selected, and so the distributions are plotted for all the leaves.
        The leaves are ranked following a criterion set via the argument rank_leaves_by.

        The features are sorted by feature importance in the Error Tree. The more important a feature is, the more correlated with the errors it is.
        The number of feature distributions to plot is set via top_k_features.

        Args:
            leaf_selector (None, int or array-like): the leaves whose information will be returned
                * int: Only plot the feature distributions for the leaf matching the id
                * array-like of int: Only plot the feature distributions for the leaves matching the ids
                * None (default): Plot the feature distributions for all the leaves

            top_k_features (int): Number of features to plot per node.
                * If a positive integer k is given, the distributions of the first k features (first in the sense of their importance) are plotted
                * If a negative integer k is given, the distributions of all but the k last features (last in the sense of their importance) are plotted
                * If k is 0, all the feature distributions are plotted

            show_global (bool): Whether to plot the feature distributions for the whole data (global baseline) along with the ones for the leaf samples.

            show_class (bool): Whether to show the proportion of Wrongly and Correctly predicted samples for each bin.

            rank_leaves_by (str): Ranking criterion for the leaves. Valid values are:
                * 'total_error_fraction': rank by the fraction of total error in the node
                * 'purity': rank by the purity (ratio of wrongly predicted samples over the total number of node samples)
                * 'class_difference': rank by the difference of number of wrongly and correctly predicted samples
                in a node.

            nr_bins (int): Number of bins in the feature distribution plots. Defaults to 10.

            figsize (tuple of float): Tuple of size 2 for the size of the plots as (width, height) in inches. Defaults to (15, 10).

        """
        ranked_feature_ids = self._error_analyzer.pipeline_preprocessor.get_top_ranked_feature_ids(self._error_clf.feature_importances_, top_k_features)

        x = self._error_analyzer.pipeline_preprocessor.inverse_transform(self._error_analyzer._error_train_x)[:, ranked_feature_ids]
        y = self._error_analyzer._error_train_y
        feature_names = self._original_feature_names

        min_values, max_values = x.min(axis=0), x.max(axis=0)
        total_error_fraction_sample_ids = y == ErrorAnalyzerConstants.WRONG_PREDICTION
        nr_wrong = self._error_clf.tree_.value[:, 0, self._error_tree.error_class_idx]

        leaf_nodes = self._get_ranked_leaf_ids(leaf_selector, rank_leaves_by)
        for leaf in leaf_nodes:
            leaf_sample_ids = self._train_leaf_ids == leaf
            nr_leaf_samples = self._error_clf.tree_.n_node_samples[leaf]
            proba_wrong_leaf = nr_wrong[leaf] / nr_leaf_samples
            suptitle = 'Leaf {} (Wrong prediction: {},'.format(leaf, format_float(proba_wrong_leaf, 3))
            suptitle += ' Correct prediction: {})'.format(format_float(1 - proba_wrong_leaf, 3))

            for i, feature_idx in enumerate(ranked_feature_ids):
                feature_name = feature_names[feature_idx]
                # TODO: use self._numerical_feature_names instead
                feature_is_numerical = not self._error_analyzer.pipeline_preprocessor.is_categorical(feature_idx)

                feature_column = x[:, i]

                if feature_is_numerical:
                    bins = np.round(np.linspace(min_values[i], max_values[i], nr_bins + 1), 2)
                    if show_class:
                        histogram_func = lambda f_samples: np.histogram(f_samples, bins=bins, density=False)[0]
                    else:
                        histogram_func = lambda f_samples: np.histogram(f_samples, bins=bins, density=True)[0]

                else:
                    bins = np.unique(feature_column)[:nr_bins]
                    if show_class:
                        histogram_func = lambda f_samples: np.bincount(np.searchsorted(bins, f_samples), minlength=len(bins))[:nr_bins]
                    else:
                        histogram_func = lambda f_samples: np.bincount(np.searchsorted(bins, f_samples), minlength=len(bins))[:nr_bins] / len(f_samples)

                if show_global:
                    if show_class:
                        hist_wrong = histogram_func(feature_column[total_error_fraction_sample_ids])
                        hist_correct = histogram_func(feature_column[~total_error_fraction_sample_ids])
                        n_samples = np.sum(hist_wrong + hist_correct)
                        normalized_hist_wrong = hist_wrong / n_samples
                        normalized_hist_correct = hist_correct / n_samples
                        root_hist_data = {
                            ErrorAnalyzerConstants.WRONG_PREDICTION: normalized_hist_wrong,
                            ErrorAnalyzerConstants.CORRECT_PREDICTION: normalized_hist_correct
                        }
                    else:
                        root_prediction = ErrorAnalyzerConstants.CORRECT_PREDICTION if nr_wrong[0] < self._error_tree.n_total_errors / 2 else ErrorAnalyzerConstants.WRONG_PREDICTION
                        root_hist_data = {root_prediction: histogram_func(feature_column)}
                else:
                    root_hist_data = None

                if show_class:
                    hist_wrong = histogram_func(feature_column[leaf_sample_ids & total_error_fraction_sample_ids])
                    hist_correct = histogram_func(feature_column[leaf_sample_ids & ~total_error_fraction_sample_ids])
                    n_samples = np.sum(hist_wrong + hist_correct)
                    normalized_hist_wrong = hist_wrong / n_samples
                    normalized_hist_correct = hist_correct / n_samples
                    leaf_hist_data = {
                        ErrorAnalyzerConstants.WRONG_PREDICTION: normalized_hist_wrong,
                        ErrorAnalyzerConstants.CORRECT_PREDICTION: normalized_hist_correct
                    }
                else:
                    leaf_prediction = ErrorAnalyzerConstants.CORRECT_PREDICTION if proba_wrong_leaf < .5 else ErrorAnalyzerConstants.WRONG_PREDICTION
                    leaf_hist_data = {leaf_prediction: histogram_func(feature_column[leaf_sample_ids])}

                x_ticks = range(len(bins))
                _BaseErrorVisualizer._add_new_plot(figsize, bins, x_ticks, feature_name, suptitle)
                _BaseErrorVisualizer._plot_feature_distribution(x_ticks, feature_is_numerical, leaf_hist_data, root_hist_data)

        plt.show()
